﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2019 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [Item("Constraint Satisfaction Evaluator", "Basic evaluator which checks for constraint satisfaction and returns a soft penalty.")]
  [StorableType("1CFBDAD0-1059-4908-9A76-C02976C367E6")]
  public class SymbolicRegressionConstraintSatisfactionEvaluator : SymbolicRegressionSingleObjectiveEvaluator {
    private const string UseBinaryPenaltyValuesParameterName = "UseBinaryPenaltyValues";
    private const string IntervalConstraintParameterName = "Constraint";

    private readonly IntervalInterpreter interpreter = new IntervalInterpreter();
    private IntervalConstraint constraint = null;

    public IFixedValueParameter<StringValue> IntervalConstraintParameter {
      get { return (IFixedValueParameter<StringValue>)Parameters[IntervalConstraintParameterName]; }
    }

    public IFixedValueParameter<BoolValue> UseBinaryPenaltyValuesParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[UseBinaryPenaltyValuesParameterName]; }
    }

    public bool UseBinaryPenaltyValues {
      get { return UseBinaryPenaltyValuesParameter.Value.Value; }
      set { UseBinaryPenaltyValuesParameter.Value.Value = value; }
    }

    public string IntervalConstraint {
      get { return IntervalConstraintParameter.Value.Value; }
      set { IntervalConstraintParameter.Value.Value = value; }
    }

    public SymbolicRegressionConstraintSatisfactionEvaluator() {
      Parameters.Add(new FixedValueParameter<BoolValue>(UseBinaryPenaltyValuesParameterName));
      Parameters.Add(new FixedValueParameter<StringValue>(IntervalConstraintParameterName));

      IntervalConstraintParameter.Value.ValueChanged += (sender, args) => constraint = IntervalConstraintsParser.Parse(IntervalConstraint).FirstOrDefault();
    }

    protected SymbolicRegressionConstraintSatisfactionEvaluator(SymbolicRegressionConstraintSatisfactionEvaluator original, Cloner cloner) : base(original, cloner) { }

    [StorableConstructor]
    protected SymbolicRegressionConstraintSatisfactionEvaluator(StorableConstructorFlag _) : base(_) { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(UseBinaryPenaltyValuesParameterName)) {
        Parameters.Add(new FixedValueParameter<BoolValue>(UseBinaryPenaltyValuesParameterName));
      }
    }

    public override bool Maximization => false;

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionConstraintSatisfactionEvaluator(this, cloner);
    }

    public override IOperation InstrumentedApply() {
      var tree = SymbolicExpressionTreeParameter.ActualValue;
      var problemData = ProblemDataParameter.ActualValue;

      double quality = Evaluate(this.ExecutionContext, tree, problemData, problemData.TrainingIndices);

      QualityParameter.ActualValue = new DoubleValue(quality);
      return base.InstrumentedApply();
    }

    public override double Evaluate(IExecutionContext context, ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows) {
      constraint = constraint ?? IntervalConstraintsParser.Parse(IntervalConstraint).FirstOrDefault();

      ISymbolicExpressionTree t = tree;
      if (constraint.IsDerivation) {
        var order = constraint.NumberOfDerivation;
        while (order-- > 0) {
          t = DerivativeCalculator.Derive(t, constraint.Variable);
        }
      }
      const double maxPenalty = double.MaxValue; // TODO: find an appropriate value for this
      var satisfied = CheckConstraintSatisfaction(tree, problemData, rows, interpreter, constraint, out Interval interval, out double lowerBoundPenalty, out double upperBoundPenalty);

      if (satisfied) return 0;
      if (UseBinaryPenaltyValues) return 1;

      var penalty = lowerBoundPenalty + upperBoundPenalty;

      if (double.IsNaN(penalty) || double.IsInfinity(penalty) || penalty > maxPenalty) {
        penalty = maxPenalty;
      }

      return penalty;
    }

    public static bool CheckConstraintSatisfaction(ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows, IntervalInterpreter interpreter, IntervalConstraint constraint, out Interval interval, out double lowerBoundPenalty, out double upperBoundPenalty) {
      ISymbolicExpressionTree t = tree;
      if (constraint.IsDerivation) {
        var order = constraint.NumberOfDerivation;
        while (order-- > 0) {
          t = DerivativeCalculator.Derive(t, constraint.Variable);
        }
      }
      interval = interpreter.GetSymbolicExpressionTreeInterval(t, problemData.Dataset, rows);
      lowerBoundPenalty = Math.Max(constraint.Interval.LowerBound - interval.LowerBound, 0);
      upperBoundPenalty = Math.Max(interval.UpperBound - constraint.Interval.UpperBound, 0);
      return constraint.Interval.Contains(interval, constraint.InclusiveLowerBound, constraint.InclusiveUpperBound);
    }
  }
}
